/* HTTP server
 * responds to GET / HTTP/1.0 with root.htm
 * responds to GET /0c:/filename.ext|/ HTTP/1.0 with the file
 * no binary transfers or searches
 * based entirely on Chris McNeil's Gopher server
 * ashok@biochemistry.cwru.edu
 *
 * Maintain Gopher item types.  Allows to add search and binary
 * file retrieval later on.
 */

/* Above comments of Ashok don't apply anymore.
   responds to GET /path/filename.ext HTTP/1.0 or,
	       GET /drive:/path/filename.ext HTTP/1.0
   If it's a directory, responds with welcome.htm or
   a directory menu is prepared on the fly.
   No searches, but, everything is transfered in binary mode.

   Selcuk Ozturk // seost2+@pitt.edu // March 31, 1995
 */

/* UMN's wonderful Gopher Server */
/* Coded by Chris McNeil, Mount Allison University */
/* cmcneil@mta.ca  */
/* Parts of code taken from Fingerd.c  */


/* Copyright 1991 Phil Karn, KA9Q */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <dir.h>
#include <dos.h>
#include <ctype.h>
#include "global.h"
#ifdef HTTP
#include "files.h"
#include "mbuf.h"
#include "socket.h"
#include "session.h"
#include "proc.h"
#include "dirutil.h"
#include "commands.h"
#include "mailbox.h"
#include "netuser.h"
#include "ftp.h"
#include "smtp.h"
#include "alloc.h"
#include "cmdparse.h"


int HttpUsers;
static char Statdir[] = ":/wwwstat/";
static char entry[] = "%-6s<a href=\"/%c:%s%s\">%13s</a>%15ld bytes   %2d %s 19%d\n";

static char Ok[] = "HTTP/1.0 200\n";
static char Nosupp[] = "HTTP/1.0 400\n\n<TITLE>Error</TITLE>\n<H1>Error 400</H1>\nUnknown method.\n";

static char *Hdrs[] = {
  "Content-Type:",
  "Content-Length:"
};

static char *CTypes1[] = {
  "video/",
  "text/",
  "image/",
  "audio/",
  "application/"
};

static char *CTypes2[] = {
  "x-msvideo",
  "quicktime",
  "quicktime",
  "mpeg",
  "mpeg",
  "richtext",
  "html",
  "x-xpixmap",
  "x-xbitmap",
  "x-rgb",
  "x-portable-pixmap",
  "x-portable-graymap",
  "x-portable-bitmap",
  "x-portable-anymap",
  "x-cmu-raster",
  "tiff",
  "jpeg",
  "jpeg",
  "gif",
  "x-wav",
  "x-aiff",
  "basic",
  "basic",
  "x-tar",
  "x-trof-man",
  "rtf",
  "postscript",
  "postscript",
  "x-stuffit",
  "mac-binhex40",
  "fractals",
  "x-zip",
  "x-gzip",
  "x-compress",
  "octet-stream",
  "plain"
};

static char *Ext[] = {
  "avi",
  "qt",
  "mov",
  "mpeg",
  "mpg",
  "rtx",
  "html",
  "xpm",
  "xbm",
  "rgb",
  "ppm",
  "pgm",
  "pbm",
  "pnm",
  "ras",
  "tiff",
  "jpeg",
  "jpg",
  "gif",
  "wav",
  "aiff",
  "au",
  "snd",
  "tar",
  "man",
  "rtf",
  "eps",
  "ps",
  "sit",
  "hqx",
  "fif",
  "zip",
  "gz",
  "z",
  NULLCHAR
};


/* function prototypes */
static void http __ARGS((int s,void *unused,void *p));
static int mkwelcome __ARGS((int s,char *Hdir,char *file));
static int dohttptdisc __ARGS((int argc,char *argv[],void *p));
/* static int dohttpsim __ARGS((int argc,char *argv[],void *p)); */
static int dohttpmax __ARGS((int argc,char *argv[],void *p));
static int dohttpdrive __ARGS((int argc,char *argv[],void *p));
extern char *decode(char *string);
extern int findport(int s, char *dir, char c);

static struct cmds Httpcmds[] = {
  "drive",       dohttpdrive,   0, 0, NULLCHAR,
  "maxcli",      dohttpmax,     0, 0, NULLCHAR,
  "tdisc",       dohttptdisc,   0, 0, NULLCHAR,
  NULLCHAR,
};

int
dohttp(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return subcmd(Httpcmds,argc,argv,p);
}


static char DR = 'c';

static int
dohttpdrive(argc,argv,p)
int argc;
char *argv[];
void *p;
{
  if(argc < 2)
    tprintf("Def. drive: %c\n",DR);
  else if(!isalpha(*argv[1]))
    tputs("Bad drive letter\n");
  else
    DR = *argv[1];
  return 0;
}

int Httpmax = 10;

static int
dohttpmax(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setint(&Httpmax,"Max. connects",argc,argv);
}
/*
int Httpsimc = 5;

static int
dohttpsim(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setint(&Httpsimc,"Simult. conn.'s serviced",argc,argv);
}
*/
int Httptdisc = 180;

static int
dohttptdisc(argc,argv,p)
int argc;
char *argv[];
void *p;
{
    return setint(&Httptdisc,"Server tdisc (sec)",argc,argv);
}

void send_scanning(FILE *fp, int s, char *buf, int buflen)
{
    unsigned long count = 0;
    time_t t;
    struct stat ft;
    char *cp1, *cp2, c, cnt[15];
    FILE *tfp;
    
    sockmode(s,SOCK_BINARY);
    
    while((fgets(buf,buflen,fp)) != NULLCHAR){
        pwait(NULL);
        if((cp1 = strstr(buf,"<SERVER")) == NULLCHAR) {
            usputs(s,buf);
        } else {
            cp1 = skipnonwhite(cp1);
            cp1 = skipwhite(cp1);
            c = tolower(*cp1);
            switch(c) {
            case 'c':

            case 'i':
            case 's':
                cp1 = strchr(cp1,'=');
                if(cp1 == NULLCHAR)
                    break;
                cp1++;
                cp2 = strchr(cp1,'>');
                if(cp2 == NULLCHAR)
                    break;
                *cp2 = '\0';
                cp2 = strdup(cp1);
                sprintf(buf,"%c%s%s",DR,Statdir,cp2);
                free(cp2);
                if((tfp = fopen(buf,READ_TEXT)) != NULLFILE) {
                    fgets(cnt,sizeof(cnt),tfp);
                    count = atol(cnt);
                    fclose(tfp);
                }
                if(c == 'i')
                    count++;
                if(c != 's')
                    usprintf(s,"%ld ", count);
                if(c != 'c' && (tfp = fopen(buf,WRITE_TEXT)) != NULLFILE) {
                    fprintf(tfp,"%ld",count);
                    fclose(tfp);
                }
                break;
            case 't':
                sprintf(buf,"%c%stcount.dat",DR,Statdir);
                if((tfp = fopen(buf,READ_TEXT)) != NULLFILE) {
                    fgets(buf,buflen,tfp);
                    fclose(tfp);
                    count = atol(buf);
                }
                usprintf(s,"%ld ",count);
                break;
            case 'd':
                time(&t);
                cp1 = ptime(&t);
                rip(cp1);
                usputs(s,cp1);
                break;
            case 'u':
                fstat(fileno(fp),&ft);
                cp1 = ptime(&ft.st_mtime);
                rip(cp1);
                usputs(s,cp1);
            }
        }
    }
}
/* Start up http service */
int httpstart(int argc, char *argv[], void *p)
{

	int16 port;

	if(argc < 2)
	  port = IPPORT_HTTP;
	else
	  port = atoi(argv[1]);

	return start_tcp(port,"HTTP Server",http,1560);

      }

static void http(int s, void *unused, void *p)
{
	struct sockaddr fsocket;
	char command[PLINELEN], *cp, *url, *tmp;
	FILE *fp;
	char file[PLINELEN+16],Hdir[SLINELEN];
	int Ctype1, port, length, i, err = 1, Nohdr = 0;
	struct stat buf;
        unsigned long count = 0;

	sockowner(s, Curproc);
	sockmode(s, SOCK_ASCII);
	HttpUsers++;

	if(availmem() < Memthresh || Httpmax < HttpUsers) {
	  usprintf(s,"HTTP/1.0 502\n%s %s%s\n\n<TITLE>Busy</TITLE><H1>Busy</H1>\nTry again later.\n",Hdrs[0],CTypes1[1],CTypes2[6]);
	  goto quit;
	}
/*
	while(HttpUsers > Httpsimc)
	  pwait(&HttpUsers);
*/
	alarm(Httptdisc * 1000L);
	if(recvline(s, command, PLINELEN) == -1) {
	  alarm(0L);
	  shutdown(s,2);
	  goto quit;
	}
	alarm(0L);

	port = findport(s,Hdir,'w');

	rip(command);
	tmp = decode(command);
	log(s,"HTTP #%u: %s",port,command);

	if(strstr(tmp,"HTTP") == NULLCHAR)
	  Nohdr = 1;

	url = skipnonwhite(tmp);
	*url = '\0';
	url++;
	url = skipwhite(url);
	url++;
	cp = skipnonwhite(url);
	*cp = '\0';
	length = strlen(url);

	if(strcmp(tmp,"GET") != 0) {
	  usprintf(s,Nosupp,tmp);
	  free(tmp);
	  goto quit;
	}

	if(!strnicmp(url,"b:",2))
	  url[0] = DR;

	if((length == 0) || !strcmp(&url[1],":/")) {
	  sprintf(file,"%c:%s/root.htm",(url[1] == ':') ? url[0] : DR,Hdir);
	} else if(url[1] == ':') {
	  sprintf(file,"%c:%s%s",url[0],Hdir,&url[2]);
	} else {
	  sprintf(file,"%c:%s/%s",DR,Hdir,url);
	}

	/* delete any .. to avoid */
	/* anyone telnetting to http and cd'ing to a directory */
	/* other than Httpdir */

	for ( i=0; i <= strlen(file); i++ ) {
	  if (file[i] == '.' && file[i+1] == '.') {
	    file[i] = ' ';
	    file[i+1] = ' ';
	  }
	}

	if(file[strlen(file)-1] == '/') {
	  strcat(file,"welcome.nhd");
	  if((fp = fopen(file,READ_BINARY)) != NULLFILE)
	    Nohdr = 1;
	  else {
	    cp = strrchr(file,'.');
	    *cp = '\0';
	    strcat(file,".htm");
	    if((fp = fopen(file,READ_BINARY)) == NULLFILE)
	      err = mkwelcome(s,Hdir,file);
	  }
	} else if((fp = fopen(file, READ_BINARY)) == NULLFILE &&
		  !access(file,0)) {
	  usprintf(s,"HTTP/1.0 302 Found\nLocation: http://%s/%s/\n\n",
		   Hostname,url);
	  usprintf(s,"<TITLE>Redirection</TITLE><H1>Redirection</H1>Please,"
		   "\n<A HREF=\"http://%s/%s/\">click here</A>\n",
		   Hostname,url);
	}

	if (fp != NULLFILE){
	  if(!Nohdr) {
	    if((cp = strrchr(file,'.')) != NULLCHAR) {
	      cp++;
	      for(i=0;Ext[i] != NULLCHAR;i++) {
		if(strlen(cp) < 3) {
		  if(!stricmp(Ext[i],cp))
		    break;
		} else if(!strnicmp(Ext[i],cp,strlen(cp)))
		  break;
	      }

	      if(i < 5)
		Ctype1 = 0;
	      else if(i < 7)
		Ctype1 = 1;
	      else if(i < 19)
		Ctype1 = 2;
	      else if(i < 23)
		Ctype1 = 3;
	      else if(i < 34)
		Ctype1 = 4;
	      else if(isbinary(fp))
		Ctype1 = 4;
	      else {
		Ctype1 = 1;
		i = 35;
	      }
	    } else if(isbinary(fp)) {
	      Ctype1 = 4;
	      i = 34;
	    } else {
	      Ctype1 = 1;
	      i = 35;
	    }

	    fstat(fileno(fp),&buf);
	    usputs(s,Ok);
	    usprintf(s,"%s %s%s\n%s %ld\n\n",Hdrs[0],CTypes1[Ctype1],
		     CTypes2[i],Hdrs[1],buf.st_size);
	  }
          if(i == 6) {
            fgets(command,PLINELEN,fp);
            if(strstr(command,"<SERVER SCAN>")) {
               send_scanning(fp,s,command,PLINELEN);
            } else {
               rewind(fp);
	       sendfile(fp, s, IMAGE_TYPE, 0, NULL);
            }
          } else
            sendfile(fp,s,IMAGE_TYPE,0,NULL);
	  fclose(fp);
	} else if(err)
	  usprintf(s, "HTTP/1.0 404\n\n<TITLE>Error</TITLE><H1>404 Not Found</H1>The file \"/%s\" does not exist on this server.\n",url);
	
	free(tmp);
 
quit:
	close_s(s);
        sprintf(command,"%c%stcount.dat",DR,Statdir);
        if((fp = fopen(command,READ_TEXT)) != NULLFILE) {
            fgets(file,sizeof(file),fp);
            fclose(fp);
            count = atol(file);
        }
        count++;
        if((fp = fopen(command,WRITE_TEXT)) != NULLFILE) {
            fprintf(fp,"%ld",count);
            fclose(fp);
        }
        HttpUsers--;
/*	psignal(&HttpUsers,0); */

}


int http0(int argc, char *argv[], void *p)
{
  int16 port;

  if(argc < 2)
    port = IPPORT_HTTP;
  else
    port = atoi(argv[1]);

  return stop_tcp(port);
}

static 
int mkwelcome(int s,char *Hdir,char *file)
{
  char *dirstring,*p,*cp;
  FILE *fp;                
  struct ffblk ffblk;
  int done,pre = 0;

  if((dirstring = (char *)mallocw(strlen(file) + 16)) == NULL)
    return 0;

  strcpy(dirstring,file);
  cp = strrchr(dirstring,'/'); 
  *++cp = '\0';

  usputs(s,Ok);
  usprintf(s,"%s %s%s\n\n",Hdrs[0],CTypes1[1],CTypes2[6]);
  strcat(dirstring,"wwwpre.cat");
  if((fp = fopen(dirstring,READ_BINARY)) != NULLFILE) {
    sendfile(fp,s,IMAGE_TYPE,0,NULL);
    fclose(fp);
    pre = 1;
  }

  *cp = '\0';
  strcat(dirstring,"*.*");
  done = findfirst(dirstring,&ffblk,FA_DIREC);

  sockmode(s,SOCK_ASCII);
  setflush(s,-1); /* we will do our own flushing */
  *cp = '\0';
  p = dirstring + strlen(Hdir) + 2;
  undosify(p);
  if(!pre)
    usprintf(s,"<TITLE>Dir</TITLE><H2>Directory of \"%c:%s\"</H2>\n",file[0],p);
  usputs(s,"<PRE>\n");

  while(!done) {
    strlwr(ffblk.ff_name);
    if(ffblk.ff_name[0] != '.') {
      if(ffblk.ff_attrib == FA_DIREC)
	cp = "[DIR]";
      else
	cp = "[FILE]";
      usprintf(s,entry,cp,file[0],p,ffblk.ff_name,ffblk.ff_name,ffblk.ff_fsize,
	       (ffblk.ff_fdate) & 0x1f, /* day */
	       Months[((ffblk.ff_fdate >> 5) & 0xf)-1], /* month */
	       (ffblk.ff_fdate >> 9) + 80); /* year */
    }
    pwait(NULL);
    done = findnext(&ffblk);
  }
  usputs(s,"</PRE>\n");
  usflush(s);

  strcat(dirstring,"wwwpost.cat");
  fp = fopen(dirstring,READ_BINARY);
  if(fp != NULLFILE) {
    sendfile(fp,s,IMAGE_TYPE,0,NULL);
    fclose(fp);
  }
  free(dirstring);
  return 0;
}
#endif /*HTTP*/
